#!/usr/bin/perl

# comprom.pl
# Generates compressed ROMFS ROM image from data files
# with or without .inf attribute files.
# .inf format compliant with W.Scholten's spec, v.0.83
# Greg Cook, 14/Oct/2022

# Usage:
#	perl comprom.pl [-f NUM] [-m SIZE] [-t TOKENS] [-i TITLE]
#		[-l] [-e] [-D] [-X]
#		FILE... > output.rom  2> output.rom.inf
# Switches:
#	-f NUM sets the maximum number of strings to find for any
#		given length.  Best range 64..512.
#	-m SIZE sets the maximum size of dictionary entries
#		(limit 255 characters).
#	-t TOKENS sets the number of tokens available for the
#		dictionary.  The remaining tokens are allocated to
#		digraphs.
#	-i TITLE adds a title file to the start of the stream.
#		TITLE will be bracketed with asterisks, then truncated
#		to 10 characters.
#	-l means files marked Locked will be stored locked
#		(by default all files are stored unlocked).
#	-e produces a ROM optimised for the Electron
#		(default is optimised for the BBC B).
#		ROM works in all models but the *RUN bugfix
#		is OS-specific.
#	-D turns on debugging output.
#	-X prints the -f, -m, -t settings and the compressed size
#		separated by commas instead of outputting
#		the ROM image.
# Arguments: Names of files to be included in the ROM image.
#	Each file's attributes are taken from the file with the same
#	name plus '.inf', if it exists.  The first line should
#	have this format:
#	[TAPE] <BBC name> <hex load address> <hex exec address>
#		[<hex length>] [Locked] [CRC=<hex CRC value>]
#	e.g. $.XFER51 FFFF1900 FFFF8023 174B CRC=2A90
#	The TAPE keyword causes the name and addresses to be used
#	verbatim (subject to length restrictions).
#	The .inf file for files named TAPE should start with $.TAPE or
#	TAPE TAPE.
# Standard input: None
# Standard output: ROMFS binary image ready for programming
# Standard error: .inf file for the ROMFS image
#	(useful if transferring to the BBC Micro as a file),
#	or an error message if the image will not fit a 16K ROM.
#	Debugging messages will appear here.
# Exit status: 0 = success, >0 = an error occurred, see standard error.

use Getopt::Std;
getopts('DXef:i:lm:t:');
map {$_=defined($_)} ($opt_l,$opt_e,$opt_D,$opt_X);
$opt_f ||= 128;
$opt_m = 255 if $opt_m < 1 || $opt_m > 255;
$opt_t = 31 if $opt_t < 1 || $opt_t > 63;

binmode(STDOUT);

# The ROM header and compressed ROMFS server, v.2.10.

$output	=
	 '4C 1C 81 4C 40 80 82 2D 13 43 6F 6D 70 72 65 73 '
	.'73 65 64 20 52 4F 4D 46 53 20 00 32 2E 31 30 20 '
	.'28 31 34 20 4F 63 74 20 32 30 32 32 29 00 28 43 '
	.'29 32 30 32 32 20 47 2E 43 6F 6F 6B 00 00 00 00 '
	.'AA E0 0E F0 38 E0 0D D0 64 98 20 76 80 90 5E 20 '
	.'74 80 85 F5 20 57 80 BA BD 16 01 49 F3 D0 0D BD '
	.'15 01 49 0F D0 06 85 C9 A2 C2 86 C8 AD 1E 81 AE '
	.'1F 81 D0 33 B5 E7 49 FF 29 0F C5 F4 60 20 74 80 '
	.'D0 2B A0 00 A5 F6 84 F6 A6 F7 30 31 06 C5 F0 3C '
	.'66 C5 C6 C5 10 10 AD C8 03 85 CC AD C6 03 6D C9 '
	.'03 85 CD 20 0A 81 A8 85 F6 86 F7 A2 00 8A A6 F4 '
	.'60 69 04 29 87 69 FE F0 2A 85 C5 D0 E6 8E 93 03 '
	.'8D 9D 03 84 C5 E0 09 E8 90 02 A2 09 DE 80 03 F0 '
	.'F4 20 0A 81 10 D0 C9 F4 B0 D7 9D 89 03 0A ED 1D '
	.'81 B0 11 20 0A 81 B0 CF CA 9D 94 03 20 0A 81 9D '
	.'80 03 90 DD A8 AD 20 81 85 F6 AD 21 81 20 18 81 '
	.'88 48 B1 F6 A8 68 E6 C5 D0 9D BC 94 03 BD 8A 03 '
	.'FE 94 03 D0 03 FE 8A 03 85 F7 B1 F6 60 '
	;

$output =~ tr/ //d;
$output = pack('H*',$output);

if($opt_e) {
	# insert address of *RUN in Electron OS 1.00
	substr($output,0x08,1,"\x14");
	substr($output,0x5C,1,"\xF2");
	substr($output,0x63,1,"\x65");
}

# The address of the fixed fields.
$addr = 0x811D;

# The lowest escape token (hard-coded).
$fet = 0xF4;

# Predefined tokens for a run or literal escape of length n.
# Other lengths use $run[0] or $lit[0] and follow it with a length byte.
# This length byte must not be zero and so $run[1] MUST be defined.
# &compress() also relies on $run[2].

@run[0,1,2,3,4,128] = ("\xFC","\xF5","\xFD","\xFE","\xFF","\xF4");
@lit[0,1,2,3,4,5,6,128] = ("\xFC","\xF5","\xF6","\xF7","\xF8",
	"\xF9","\xFA","\xFB");

# Add the size of the fixed fields (5) to $addr to give
# the start address of the dictionary.
# $addr will be increased by the size of the digraph table and the
# dictionary later, to point to the ROMFS stream.

$addr += 0x05;

# generate ROMFS stream

@fileset = ();

if(defined($opt_i)) {
	$name=substr('*'.$opt_i.'*',0,10);
	$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, 0xFFFFFFFF,
		0xFFFFFFFF, 0, 0, 0x80);
	push @fileset, [ [ $header, '' ] ];
}

while(($file=shift @ARGV) ne '') {
	open(FILE,'<'.$file) or die "Cannot read '$file'";
	binmode(FILE);

	# Collect particulars from the .inf file, or supply defaults.
	if(open(INF,'<'.$file.'.inf')) {
		# Read first line.
		$infline = <INF>;
		close(INF);
		# Scan for data fields.
		($tape,$name,$load,$exec,$reported_length,$lock,$crc) =
			$infline =~ /
				# optional TAPE keyword at start of line
				^((?:TAPE)?) \s*
				# 1..10 char name, discard excess
				(\S{1,10}) \S* \s+
				# 1..8 hex digits load address
				([0-9A-Fa-f]{1,8}) \s+
				# 1..8 hex digits exec address
				([0-9A-Fa-f]{1,8}) \s*
				# 1..8 hex digits length if present
				# do not select CRC= or ATTR=
				((?:(?!\S*=)[0-9A-Fa-f]{1,8})?) \s*
				# lock attribute
				((?:Locked)?) \s*
				# CRC field
				((?:CRC\s*=\s*[0-9A-Fa-f]{1,4})?)
			/x;

		# remember if these fields were filled
		$reported_length_1 = $reported_length;
		$crc=~s/^CRC\s*=\s*//;
		$crc_1=$crc;

		# convert to decimal
		($load,$exec,$reported_length,$crc) =
			map(hex,($load,$exec,$reported_length,$crc));

		unless($tape) {
			# strip default disc directory
			if(substr($name,0,2) eq '$.') {
				$name = substr($name,2);
			}

			# fixup 18- or 24-bit addresses
			# sometimes returned by *INFO
			# to 32 bit addresses as used in ROMFS
			$load &= 0x3ffff;
			if(($load & 0x30000) == 0x30000) {
				$load |= 0xFFFF0000;
			}
			$exec &= 0x3ffff;
			if(($exec & 0x30000) == 0x30000) {
				$exec |= 0xFFFF0000;
			}
		}

		# read in whole file!
		$blob = join('', <FILE>);
		close(FILE);
		$blob_length = length($blob);

		# Complain if file and .inf mismatch. If .inf doesn't
		# define length or CRC, don't test.
		if($reported_length_1 ne ''
			&& $blob_length != $reported_length) {
			die "File '$file' length ($blob_length) "
				."!= reported length ($reported_length)";
		}
		$blob_crc = &crc16($blob);
		if($crc_1 ne '' && $blob_crc != $crc) {
			die "File '$file' CRC ($blob_crc) != "
				."reported CRC ($crc)";
		}
	} else {
		# Supply default attributes for file.
		$name = substr($file,0,10);
		$load = $exec = 0xFFFFFFFF;
		$lock = '';
		# Read whole file.
		$blob = join('', <FILE>);
		close(FILE);
		$blob_length = length($blob);
	}

# @fileset is a list of files, each file a list of one or two segments,
# each segment a list containing the header and the data block(s)
# in two strings.  See perllol(1).

	@file = ();

	$block_flag  = 0x00;
	$block_flag |= 0x01 if $lock && $opt_l;
	$block = 0;

	# Process all but the last block.
	if($blob_length > 256) {
		# Generate ROMFS file header excluding unknown fields
		$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, $load,
			$exec, 0, 256, $block_flag);

		$datastr = undef;
		for($block=0; $block+256 < $blob_length; $block+=256) {
			$data = substr($blob,$block,256);
			$data .= pack('n',&crc16($data));
			if($datastr ne '') {
				$datastr .= pack('C',0x23);
			}
			$datastr .= $data;
		}

		# Inserts in @file a reference to an anonymous list
		# containing the two strings.
		push @file, [ $header, $datastr ];
	}

	# Process last or only block.
	# Generate ROMFS file header excluding unknown fields
	$block_flag |= 0x80;
	# $block_flag |= 0x40 if $blob_length == $block;
	$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, $load, $exec,
		$block/256, $blob_length-$block, $block_flag);
	$data = substr($blob,$block);
	unless($blob_length-$block == length($data)) {
		die 'assertion failed';
	}
	$data .= pack('n',&crc16($data)) unless $blob_length == $block;

	# Appends to @file a reference to an anonymous list
	# containing the two strings.
	push @file, [ $header, $data ];

	# Appends to @fileset a reference to the list currently
	# named @file, containing the reference(s).
	push @fileset, [ @file ];
}

# discard these variables
undef $blob, @file;

# flatten structure into list of strings

foreach $fileref (@fileset) {
	foreach $segref (@$fileref) {
		push @stringlist, @$segref;
	}
}

print STDERR "$0: searching for repeating strings\n" if $opt_D;

# get most common digraphs
@dictionary = @prefixes = &gram_scan(1,0,$opt_f,256,map(chr,0..255),
	@stringlist);
print STDERR "$0: found ",scalar(@prefixes)," strings of length 2\n"
	if $opt_D;

# Grow the most common strings up to 255 characters until
# they no longer repeat.

$stringlen = 2;
$dir=1;
while(@prefixes && ++$stringlen <= $opt_m) {
	# This potentially will not find strings in their entirety.
#	@prefixes = &gram_scan(0,$dir^=1,$opt_f,
#		scalar(@prefixes),@prefixes,@stringlist);

	@right = &gram_scan(0,1,$opt_f,
		scalar(@prefixes),@prefixes,@stringlist);
	@left = &gram_scan(0,0,$opt_f,
		scalar(@prefixes),@prefixes,@stringlist);
	%temp = map {($_,undef)} (@left,@right);
	# Modern Perl releases return hash keys in a nondeterministic
	# order to inhibit DDOS attack.  The order was never guaranteed
	# (see perlsec(1)), though older Perl versions maintained a
	# consistent, but still arbitrary, order between script runs.
	# Dictionary entries are now sorted during compilation to ensure
	# consistency; the changed order of the entries, combined with
	# the compression algorithm's appalling behaviour, will cause
	# the minimum achievable ROM image size for certain file sets to
	# increase, whereas for others it will decrease in equal measure.
	# Conversely, removing 'sort' from the next line makes it
	# possible to repeat a comprom.pl invocation several times
	# and keep the shortest resulting ROM image.
	@prefixes = sort keys %temp;

	unshift @dictionary, @prefixes;

	print STDERR "$0: found ",scalar(@prefixes),
		" strings of length $stringlen\n" if $opt_D && @prefixes;
}

undef %temp;

%count = ();
%dummy = ();

# compress the string list to find the substrings used most
print STDERR "$0: compressing to select dictionary entries\n" if $opt_D;
map {&compress($_,undef,\%count,\%dummy,[],[],\%dummy,\@dictionary)}
	@stringlist;

undef @stringlist;

# Find substrings appearing in all used dictionary entries.
# We are only interested in counts; entries disqualified for a
# low score don't remove the substring from the text.

@builddict = @dictionary;

while(($word = shift(@builddict)) ne '') {
	if($count{$word}) {
		&compress($word,undef,\%count,\%dummy,[],[],\%dummy,
			\@builddict);
	}
}

die 'assertion failed' if @builddict;
@topdict = ();

%level = ();
%tok = ();
$token = 0x81 + $opt_t;
$digs = $dict = '';

# Build the digraph table and dictionary.  Any entry with a positive
# score is included.

print STDERR "$0: dictionary entries:\n" if $opt_D;
print STDERR "$0: \ttoken\tcount\tlevel\tsaving\tlength\tentry\t"
	."compressed\n" if $opt_D;
foreach $word (sort {length($a) <=> length($b)
	|| $count{$b} <=> $count{$a}
	|| $a cmp $b} @dictionary) {
	next if $count{$word} < 2;
	if(length($word) == 2) {
		# Digraphs with high-order characters may expand
		# the data due to extra escapes!
		# However it would be daft not to have any.
		if($token < $fet && $count{$word} >
			($word =~ tr/0x80-0xff// ? 5 : 2)) {
			$line=sprintf("%s:\t%7.02x\t%7d\t%7d\t%7d"
				."\t%7d\t\"%s\"\t\"%s\"",$0,$token,
				$count{$word},0,$count{$word}-2,2,
				$word,$word) if $opt_D;
			$line=~tr/\t -~/./c if $opt_D;
			print STDERR $line,"\n" if $opt_D;
			$tok{$word} = chr($token++);
			# The digraph table now comes after the
			# dictionary.  Collect the digraphs for
			# later.
			# $addr += 2;
			$digs .= $word;
			unshift @builddict, $word;
			unshift @topdict, $word;
		}
	} elsif($addr < 0x8100 + ($opt_t << 8)) {
		$level = 1;
		$comp_word = &compress($word,\$level,\%dummy,\%level,
			\@lit,\@run,\%tok,\@builddict);
		if((length($comp_word)-2) * ($count{$word}-1) > 3) {
			$line=sprintf("%s:\t%7.04x\t%7d\t%7d\t%7d"
				."\t%7d\t\"%s\"\t\"%s\"",$0,$addr,
				$count{$word},$level,
				(length($comp_word)-2) *
				($count{$word}-1)-3,
				length($comp_word),$word,$comp_word)
				if $opt_D;
			$line=~tr/\t -~/./c if $opt_D;
			print STDERR $line,"\n" if $opt_D;
			$level{$word} = $level;
			$tok{$word} = pack('n',$addr);
			$addr += length($comp_word) + 1;
			$dict .= chr(&tokens($opt_t,$comp_word));
			$dict .= $comp_word;
			if($level < 9) {
				unshift @builddict, $word;
			}
			unshift @topdict, $word;
		}
	}
}

undef @builddict;
@dictionary = @topdict;
undef @topdict;

# Pack the fixed-location fields: the highest dictionary token,
# data pointer and digraph pointer.
# Follow with the dictionary and digraph table.

$dptr = $addr;
$addr += length($digs);
$output .= pack('Cv2a*a*',
	($opt_t<<1)|1,$addr,$dptr,$dict,$digs);
undef $dict;
undef $digs;

$level=1;

foreach $fileref (@fileset) {
	@uncomp = ();
	foreach $segref (@$fileref) {
		# Save the uncompressed version for crc16
		# (excluding the synchronising * which is not
		# counted)
		push @uncomp, substr($segref->[0],1);
		# compress the header and data
		$segref->[0] =
			&compress($segref->[0],\$level,\%dummy,\%level,
				\@lit,\@run,\%tok,\@dictionary);
		$segref->[1] =
			&compress($segref->[1],\$level,\%dummy,\%level,
				\@lit,\@run,\%tok,\@dictionary);
		# Increment address, adding 7 for the escape token,
		# ROMFS link field and CRC.
		$addr+=length($segref->[0])+length($segref->[1])+6;
		$addr+=length($lit[6]);
	}
	# now the address of the next file is known, append to the
	# output string filling in the link field and CRC.
	foreach $segref (@$fileref) {
		$output .= $segref->[0];
		$string = pack('V',$addr);
		$output .= pack('a*a*n',$lit[6],$string,
			&crc16((shift @uncomp).$string));
		$output .= $segref->[1];
	}
}

print "$opt_f,$opt_m,$opt_t,",$addr-0x7FFF,"\n" if $opt_X;
exit 0 if $opt_X;

# This could be moved into the loop above (fail fast)
# but we would like to give a clue of how much over we are.

if($addr>0xBFFF) {
	print STDERR "$0: Sorry - could not fit into 16K.\n";
	print STDERR "$0: ",$addr-0xBFFF," byte excess.\n";
	exit 1;
}

# Add end-of-ROM marker
$output .= pack('C',0x2B);

print $output;

# Print a convenient .inf line for the likes of Xfer
print STDERR sprintf("%-10s %8.04X %8.04X  %04X CRC=%04X\n",
	'$.ROMIMG',0xFFFF8000,0xFFFF8000,$addr-0x7FFF,&crc16($output));

exit 0;

######## Subroutines

sub gram_scan ($$$$@) {
	my ($overlap,$direction,$return,$prefixes,@list) = @_;

	# Return a list of the $return most common substrings
	# in @list[$prefixes..] such that each returned string is
	# equal to a string in @list[0..$prefixes-1] plus one
	# character. When $direction = 1 the character will come before
	# the prefix.
	# When $overlap = 0, skips over a string that has been counted
	# before continuing the search.  When $overlap = 1, continues
	# the search at the next character.  This is slower but yields
	# more strings.

	my (%prefix_no) = ();
	my (@prefix) = ();
	my (@return) = ();
	my (@table) = (undef) x $prefixes;
	my (@max) = ();
	my ($i,$j,$k,$prefix,$string,$no,$char,$maxcount,$maxno,
		$maxchar,$arpreftab);

	return(()) if $prefixes == 0 || $return == 0;

	# create lookup hash of prefixes to numbers
	# (and touch left-anchored substrings)

	for($i=0;$i<$prefixes;++$i) {
		$prefix = shift @list;
		for($j=1;$j<length($prefix);++$j) {
			$prefix_no{substr($prefix,0,$j)} ||= undef;
		}
		$prefix_no{$prefix} = $i;
		$prefix[$i] = $prefix;

		$table[$i] = [ (undef) x 256 ];
	}

	# search text for prefixes by testing strings of increasing length
	# count occurrences of characters 0..255 after prefix

	foreach $string (@list) {
		for($i=0;$i<length($string);++$i) {
			$j=1;
			while($i+$j<length($string) && exists($prefix_no
				{substr($string,$i,$j)})) {
				++$j;
			}
			if(--$j && ($no =
				$prefix_no{substr($string,$i,$j)}) ne'') {
				if($dir==0 && $i+$j<length($string)) {
					$char = ord(substr($string,
						$i+$j,1));
				} elsif($dir!=0 && $i>0) {
					$char = ord(substr($string,
						$i-1,1));
				}
				++$table[$no][$char];

				$i+=$j unless $overlap;
			}
		}
	}

	# return up to $return of the most common strings
	# consisting of a prefix plus one character
	# (discarding strings that occur once)

	$i=0; @max=(undef);
	while($i<$return && @max) {
		@max=();
		$maxcount = 1;
		for($j=0;$j<$prefixes;++$j) {
			$arpreftab=$table[$j];
			for($k=0;$k<256;++$k) {
				if($arpreftab->[$k] > $maxcount) {
					$maxcount=$arpreftab->[$k];
					@max=("$j,$k");
				} elsif($arpreftab->[$k] == $maxcount) {
					push @max,"$j,$k";
				}
			}
		}

		if($maxcount==1) {@max=();}
		$j=0;
		while($i<$return && $j<@max) {
			($maxno,$maxchar) = split /,/,$max[$j++];
			$prefix = $prefix[$maxno];
			$table[$maxno][$maxchar] = undef;
			if($dir==0) {
				$prefix .= chr($maxchar);
			} else {
				$prefix = chr($maxchar).$prefix;
			}
			push @return, $prefix;
			++$i;
		}
	}

#	foreach $tabref (@table) {@$tabref = ();}
#	@table = ();
	undef %prefix_no;
	return @return;
}

sub compress ($\$\%\%\@\@\%\@) {
	my ($string,$srlevel,$hrcount,$hrlevel,$arlit,$arrun,
		$hrtok,$ardict) = @_;

	# Returns $string, compressed with run-length encoding and
	# dictionary substitution.  The external hash %$hrlevel maps
	# dictionary strings to an externally defined level of recursion;
	# on exit the external scalar $$srlevel will contain 1 + the
	# highest level number of any dictionary string used if higher
	# than its present value, otherwise it will be untouched.
	# External arrays @$arlit and @$arrun contain tokens to escape
	# sequences of literals or encode runs of length n, respectively;
	# if an element is not defined then element [0] is used and
	# followed by a length byte (0x00..0x7f to encode a 1 to 128
	# char run, 0x80..0xff to escape 1 to 128 literals.)
	# External hash %$hrtok maps dictionary strings to tokens.
	# The dictionary is also provided as an external array @$ardict
	# and the first elements will be matched preferentially.

	my ($i,$j,$ind,$subst,$char,$counter);

	return $string if $string eq '';

	# try each dict entry against string until one matches
	$i=0; $ind=-1;
	while($i < @$ardict && ($ind = index($string,$ardict->[$i])) <0) {
		++$i;
	}

	if($ind>=0) {
		# replace with token, compress the strings on each side
		++$hrcount->{$ardict->[$i]};
		if($hrlevel->{$ardict->[$i]} + 1 > $$srlevel) {
			$$srlevel = $hrlevel->{$ardict->[$i]} + 1;
		}
		return &compress(substr($string,0,$ind),$srlevel,
			$hrcount,$hrlevel,$arlit,$arrun,$hrtok,$ardict)
			.$hrtok->{$ardict->[$i]}
			.&compress(substr($string,
			$ind+length($ardict->[$i])),$srlevel,$hrcount,
			$hrlevel,$arlit,$arrun,$hrtok,$ardict);
	} else {
		# no dict entries in string, test for runs.
		# Runs must be >=3 chars long to prevent insertion loss
		# in a literal sequence.
		$i=0;$j=0;$subst=undef;$char=undef;
		while($i<length($string) && $j-$i<2) {
			$char=substr($string,$i,1);
			$j=++$i;
			while($j<length($string)
				&& substr($string,$j,1) eq $char) {
					++$j;
				}
		}
		if($j-$i>=2) {
			# Run found. Replace with run token(s)
			# and compress the strings on each side
			--$i;
			$counter=$j-$i;
			while($counter>128) {
				$subst.=$arrun[128].$char;
				$counter-=128;
			}
			if(defined($arrun->[$counter])) {
				$subst.=$arrun->[$counter].$char;
			} else {
				$subst.=$arrun->[0]
					.chr($counter - 0x01).$char;
			}
			return &compress(substr($string,0,$i),
				$srlevel,$hrcount,$hrlevel,
				$arlit,$arrun,$hrtok,$ardict)
				.$subst
				.&compress(substr($string,$j),
				$srlevel,$hrcount,$hrlevel,
				$arlit,$arrun,$artok,$ardict);
		} else {
			# no runs or dict entries, therefore no need
			# to recurse.
			$i=0;$subst=undef;

			# repeat the following until the string is done:
			while($i<length($string)) {
				# copy characters to output until an
				# escapable entry is found.
				while($i<length($string)
					&& ord(substr($string,$i,1))
					<0x80) {
					$subst .= substr($string,$i++,1);
				}

				# If so and the string >128 chars,
				# blanket escape the first 128 to avoid
				# using a two byte sequence.
				if(length($string)-$i >=128) {
					$subst .= $arlit->[128]
						.substr($string,$i,128);
						$i+=128;

				# else find the last escapable char in
				# the string.
				} elsif($i<length($string)) {
					$j=length($string)-1;
					while($j>$i && ord(substr($string,
						$j,1))<0x80) {
						--$j;
					}
					$j = $j-$i + 1;
					unless($j>=1) {
						die 'assertion failed';
					}

					# if that is the character after
					# the first and they are the same,
					# write a 2-char run.
					if($j==2 && substr($string,$i,1)
						eq substr($string,
						$i+1,1)) {
						$subst .= $arrun->[2]
							.substr($string,
							$i,1);
						$i+=$j;

					# else blanket escape the string
					# between them.
					} elsif(defined($arlit->[$j])) {
						$subst .= $arlit->[$j]
						.substr($string,$i,$j);
						$i+=$j;
					} else {
						$subst .= $arlit->[0]
							.chr($j + 0x7F)
							.substr($string,
							$i,$j);
						$i+=$j;
					}
				}
			}
			return $subst;
		}
	}
}

sub tokens ($$) {
	my ($t,$string) = @_;
	# return number of tokens in $string

	my $tokens = 0;
	my $iter = 0;
	my $char;

	while($iter < length($string)) {
		++$tokens;
		if(($char = ord(substr($string,$iter,1))) < 0x80) {
			++$iter;
		} elsif($char >= 0xf4) {
			$char = (0x7f..0x85,0xff,0x00..0x03)[$char-0xf4];
			# variable length run/escape
			$char = ord(substr($string,++$iter,1)) until $char;
			if($char >= 0x80) {
				# 1 literal escape token + 1..128 literals
				$iter += $char - 0x7e;
			} else {
				$iter += 2; # run
			}
		} elsif($char >= 0x81 + $t) {
			++$iter; # digraph
		} else {
			$iter += 2; # dictionary token
		}
	}
	return($tokens);
}

sub crc16 ($) {
	# return Acorn-style 16-bit cyclic redundancy check
	# See BBC User Guide p.365 (B+ User Guide p.369)
	# Argument: binary string
	# Returns: CRC for argument (in decimal)

	my ($input) = @_;
	my ($iter,$char);
	my $crc = 0x0000;

	for($iter=0; $iter<length($input); ++$iter) {
		$char = unpack('C',substr($input,$iter,1));
		$crc = ($crc & 0xFF) << 8 | ($crc & 0xFF00) >> 8 ^ $char;
		$crc ^= (($crc & 0xF0) >> 4) * 0x0201;
		$crc ^= ($crc & 0x0F) * 0x1020;
	}
	return($crc);
}
