#!/usr/bin/perl

# rom.pl
# Generates ROMFS ROM image from data files
# with or without .inf attribute files.
# .inf format compliant with W.Scholten's spec, v.0.83
# Greg Cook, 13/Oct/2022

# Usage:
#	perl rom.pl [-i TITLE] [-l] [-e]
#		FILE... > output.rom 2> output.rom.inf
#
# Switches:
#	-i TITLE adds a title file to the start of the stream.
#		TITLE will be bracketed with asterisks, then truncated
#		to 10 characters.
#	-l means files marked Locked will be stored locked
#		(by default all files are stored unlocked).
#	-e produces a ROM optimised for the Electron
#		(default is optimised for the BBC B).
#		ROM works in all models but the *RUN bugfix
#		is OS-specific.
# Arguments: Names of files to be included in the ROM image.
#	Each file's attributes are taken from the file with the same
#	name plus '.inf', if it exists.  The first line should
#	have this format:
#	[TAPE] <BBC name> <hex load address> <hex exec address>
#		[<hex length>] [Locked] [CRC=<hex CRC value>]
#	e.g. $.XFER51 FFFF1900 FFFF8023 174B CRC=2A90
#	The TAPE keyword causes the name and addresses to be used
#	verbatim (subject to length restrictions).
#	The .inf file for files named TAPE should start with $.TAPE or
#	TAPE TAPE.
# Standard input: None
# Standard output: ROMFS binary image ready for programming
# Standard error: .inf file for the ROMFS image
#	(useful if transferring to the BBC Micro as a file),
#	or an error message if the image will not fit a 16K ROM.
#	Debugging messages will appear here.
# Exit status: 0 = success, >0 = an error occurred, see standard error.

use Getopt::Std;
getopts('ei:l');
map {$_=defined($_)} ($opt_e,$opt_l);

binmode(STDOUT);

# The ROM header and ROMFS server, v.0.10.
# Contains fixes for *RUN and OSFILE return bugs.

$output	=
	 '4C 92 80 4C 32 80 82 22 0A 52 4F 4D 46 53 20 00 '
	.'30 2E 31 30 20 28 31 33 20 4F 63 74 20 32 30 32 '
	.'32 29 00 28 43 29 32 30 32 32 20 47 2E 43 6F 6F '
	.'6B 00 AA E0 0E F0 3A E0 0D D0 54 98 20 6A 80 90 '
	.'4E 20 68 80 85 F5 A9 93 85 F6 A9 80 85 F7 20 51 '
	.'80 BA BD 16 01 49 F3 D0 34 BD 15 01 49 0F D0 2D '
	.'85 C9 A2 C2 86 C8 D0 28 B5 E7 49 FF 29 0F C5 F4 '
	.'60 20 68 80 D0 19 A1 E8 E6 F6 D0 02 E6 F7 A8 AD '
	.'C8 03 85 CC AD C6 03 18 6D C9 03 85 CD A2 00 8A '
	.'A6 F4 60 '
	;

$output =~ tr/ //d;
$output = pack('H*',$output);

if($opt_e) {
	# insert address of *RUN in Electron OS 1.00
	substr($output,0x08,1,"\x0B");
	substr($output,0x56,1,"\xF2");
	substr($output,0x5D,1,"\x65");
}

$addr = 0x8093;

# generate ROMFS stream

@fileset = ();

if(defined($opt_i)) {
	$name=substr('*'.$opt_i.'*',0,10);
	$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, 0xFFFFFFFF,
		0xFFFFFFFF, 0, 0, 0x80);
	push @fileset, [ [ $header, '' ] ];
}

while(($file=shift @ARGV) ne '') {
	open(FILE,'<'.$file) or die "Cannot read '$file'";
	binmode(FILE);

	# Collect particulars from the .inf file, or supply defaults.
	if(open(INF,'<'.$file.'.inf')) {
		# Read first line.
		$infline = <INF>;
		close(INF);
		# Scan for data fields.
		($tape,$name,$load,$exec,$reported_length,$lock,$crc) =
			$infline =~ /
				# optional TAPE keyword at start of line
				^((?:TAPE)?) \s*
				# 1..10 char name, discard excess
				(\S{1,10}) \S* \s+
				# 1..8 hex digits load address
				([0-9A-Fa-f]{1,8}) \s+
				# 1..8 hex digits exec address
				([0-9A-Fa-f]{1,8}) \s*
				# 1..8 hex digits length if present
				# do not select CRC= or ATTR=
				((?:(?!\S*=)[0-9A-Fa-f]{1,8})?) \s*
				# lock attribute
				((?:Locked)?) \s*
				# CRC field
				((?:CRC\s*=\s*[0-9A-Fa-f]{1,4})?)
			/x;

		# remember if these fields were filled
		$reported_length_1 = $reported_length;
		$crc=~s/^CRC\s*=\s*//;
		$crc_1=$crc;

		# convert to decimal
		($load,$exec,$reported_length,$crc) =
			map(hex,($load,$exec,$reported_length,$crc));

		unless($tape) {
			# strip default disc directory
			if(substr($name,0,2) eq '$.') {
				$name = substr($name,2);
			}

			# fixup 18- or 24-bit addresses
			# sometimes returned by *INFO
			# to 32 bit addresses as used in ROMFS
			$load &= 0x3ffff;
			if(($load & 0x30000) == 0x30000) {
				$load |= 0xFFFF0000;
			}
			$exec &= 0x3ffff;
			if(($exec & 0x30000) == 0x30000) {
				$exec |= 0xFFFF0000;
			}
		}

		# read in whole file!
		$blob = join('', <FILE>);
		close(FILE);
		$blob_length = length($blob);

		# Complain if file and .inf mismatch. If .inf doesn't
		# define length or CRC, don't test.
		if($reported_length_1 ne ''
			&& $blob_length != $reported_length) {
			die "File '$file' length ($blob_length) "
				."!= reported length ($reported_length)";
		}
		$blob_crc = &crc16($blob);
		if($crc_1 ne '' && $blob_crc != $crc) {
			die "File '$file' CRC ($blob_crc) != "
				."reported CRC ($crc)";
		}
	} else {
		# Supply default attributes for file.
		$name = substr($file,0,10);
		$load = $exec = 0xFFFFFFFF;
		$lock = '';
		# Read whole file.
		$blob = join('', <FILE>);
		close(FILE);
		$blob_length = length($blob);
	}

# @fileset is a list of files, each file a list of one or two segments,
# each segment a list containing the header and the data block(s)
# in two strings.  See perllol(1).

	@file = ();

	$block_flag  = 0x00;
	$block_flag |= 0x01 if $lock && $opt_l;
	$block = 0;

	# Process all but the last block.
	if($blob_length > 256) {
		# Generate ROMFS file header excluding unknown fields
		$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, $load,
			$exec, 0, 256, $block_flag);

		$datastr = undef;
		for($block=0; $block+256 < $blob_length; $block+=256) {
			$data = substr($blob,$block,256);
			$data .= pack('n',&crc16($data));
			if($datastr ne '') {
				$datastr .= pack('C',0x23);
			}
			$datastr .= $data;
		}

		# Inserts in @file a reference to an anonymous list
		# containing the two strings.
		push @file, [ $header, $datastr ];
	}

	# Process last or only block.
	# Generate ROMFS file header excluding unknowns
	$block_flag |= 0x80;
	# $block_flag |= 0x40 if $blob_length == $block;
	$header = pack('Ca*CV2v2C', 0x2A, $name, 0x00, $load, $exec,
		$block/256, $blob_length-$block, $block_flag);
	$data = substr($blob,$block);
	unless($blob_length-$block == length($data)) {
		die 'assertion failed';
	}
	$data .= pack('n',&crc16($data)) unless $blob_length == $block;

	# Appends to @file a reference to an anonymous list
	# containing the two strings.
	push @file, [ $header, $data ];

	# Appends to @fileset a reference to the list currently
	# named @file, containing the reference(s).
	push @fileset, [ @file ];
}

# discard these variables
undef $blob, @file;

foreach $fileref (@fileset) {
	foreach $segref (@$fileref) {
		# Increment address, adding 6 for the
		# ROMFS link field and CRC.
		$addr+=length($segref->[0])+length($segref->[1])+6;
	}
	# now the address of the next file is known, append to the
	# output string filling in the link field and CRC.
	foreach $segref (@$fileref) {
		$output .= $segref->[0];
		$string = pack('V',$addr);
		$output .= pack('a*n',$string,
			&crc16(substr($segref->[0],1).$string));
		$output .= $segref->[1];
	}
}

# This could be moved into the loop above (fail fast)
# but we would like to give a clue of how much over we are.

if($addr>0xBFFF) {
	print STDERR "$0: Sorry - could not fit into 16K.\n";
	print STDERR "$0: ",$addr-0xBFFF," byte excess.\n";
	exit 1;
}

# Add end-of-ROM marker
$output .= pack('C',0x2B);

print $output;

# Print a convenient .inf line for the likes of Xfer
print STDERR sprintf("%-10s %8.04X %8.04X  %04X CRC=%04X\n",
	'$.ROMIMG',0xFFFF8000,0xFFFF8000,$addr-0x7FFF,&crc16($output));

exit 0;

######## Subroutines

sub crc16 ($) {
	# return Acorn-style 16-bit cyclic redundancy check
	# See BBC User Guide p.365 (B+ User Guide p.369)
	# Argument: binary string
	# Returns: CRC for argument (in decimal)

	my ($input) = @_;
	my ($iter,$char);
	my $crc = 0x0000;

	for($iter=0; $iter<length($input); ++$iter) {
		$char = unpack('C',substr($input,$iter,1));
		$crc = ($crc & 0xFF) << 8 | ($crc & 0xFF00) >> 8 ^ $char;
		$crc ^= (($crc & 0xF0) >> 4) * 0x0201;
		$crc ^= ($crc & 0x0F) * 0x1020;
	}
	return($crc);
}
